// imports
use std::collections::HashMap;
use std::io::{self, Seek, Read, BufWriter, SeekFrom, Write, BufReader};
// use std::thread::__FastLocalKeyInner;
use std::{path::Path,
        fs::{File, 
            OpenOptions},
    };
use crc::{Crc, CRC_32_CKSUM};

use byteorder::{LittleEndian, ReadBytesExt, WriteBytesExt};
use serde_derive::{Serialize, Deserialize};


// types declarations
type ByteStr = [u8];
type ByteString = Vec<u8>;


// define the key - value pair
#[derive(Debug, Serialize, Deserialize)]
pub struct KeyValuePair {
    pub key: ByteString,
    pub value: ByteString,
}

impl KeyValuePair {

}

#[derive(Debug)]
pub struct Action {
    // variables come here
    f: File,
    index: HashMap<ByteString, u64>,

}

impl Action {

    pub fn open(path: &Path) -> io::Result<Self> {
        /* File open operation performed here */
        let f = OpenOptions::new()
                        .read(true)
                        .write(true)
                        .append(true)
                        .create(true)
                        .open(path)?;

        let index: HashMap<ByteString, u64> = HashMap::new();

        Ok(Action{f, index})
    }

    pub fn load(&mut self) -> io::Result<()> {
        /* read the file */
        let mut f = io::BufReader::new(&mut self.f);

        // update the index

        loop {
            let position = f.seek(io::SeekFrom::Current(0))?;

            let maybe_kv = Self::process_records(&mut f);
            let kv = match maybe_kv {
                Ok(kv) => kv,
                Err(err) => {
                    match err.kind() {
                        io::ErrorKind::UnexpectedEof => {
                            break;
                        },
                        _ => return Err(err),
                    }
                },
            };
            // add key to the index
            self.index.insert(kv.key, position);
        }
        Ok(())
    }

    pub fn process_records<R: Read>(f: &mut R) -> io::Result<KeyValuePair>{

        let saved_checksum = f.read_u32::<LittleEndian>()?;
        let key_len = f.read_u32::<LittleEndian>()?;
        let val_len = f.read_u32::<LittleEndian>()?;

        let data_len = key_len + val_len;
        // define the data vector
        let mut data = ByteString::with_capacity(data_len as usize); // cast to u64

        // read the data
        {
            f.by_ref()
            .take(data_len as u64)
            .read_to_end(&mut data)?;
        }
        
        // validataion
        debug_assert_eq!(data.len(), data_len as usize);

        // check checksum
        let data_checksum = Crc::<u32>::new(&crc::CRC_32_CKSUM)
                                .checksum(&data);

        if data_checksum != saved_checksum {
            panic!(
                "data corruption encountered ( {:08x} != {:08x} )",
                data_checksum, saved_checksum
            );
        }

        // split the key and value
        let value = data.split_off(key_len as usize);
        let key = data;

        Ok(KeyValuePair { key, value })
    }

    pub fn seek_to_end(&mut self) -> io::Result<u64> {
        self.f.seek(SeekFrom::End(0))
    }

    pub fn get(&mut self, key: &ByteStr) -> io::Result<Option<ByteString>> {

        let position: u64 = match self.index.get(key) {
            None => return Ok(None),
            Some(position) => *position,
        };

        let kv = self.get_at(position)?;
        Ok(Some(kv.value))

    }

    pub fn get_at(&mut self, position: u64) -> io::Result<KeyValuePair> {
        let mut f = BufReader::new(&mut self.f);
        f.seek(SeekFrom::Start(position))?;
        let kv = Self::process_records(&mut f)?;
        Ok(kv)
    }

    #[inline]
    pub fn delete(&mut self, key: &ByteStr) -> io::Result<()> {
        self.insert(key, b"")
    }

    pub fn insert(&mut self, key: &ByteStr, value: &ByteStr) -> io::Result<()> {
        
        let position = self.insert_ignore_index(key, value)?;
        
        self.index.insert(key.to_vec(), position);

        Ok(())
    }

    pub fn insert_ignore_index(&mut self, key: &ByteStr, 
            value: &ByteStr) -> io::Result<u64> {
        
        let mut f = BufWriter::new(&mut self.f);
        let key_len = key.len();
        let value_len = value.len();
        let mut data_tmp = ByteString::with_capacity(key_len + value_len);
        
        // joining the iterator to make one loop
        for byte in key.into_iter().chain(value.into_iter()) {
            data_tmp.push(*byte);
        }

        let checksum: u32 = Crc::<u32>::new(&CRC_32_CKSUM)
                                .checksum(&data_tmp);

        // move the pointer to end of the file
        let next_byte = SeekFrom::End(0);
        let current_position = f.seek(SeekFrom::Current(0))?;

        // write to file
        f.seek(next_byte)?;
        f.write_u32::<LittleEndian>(checksum)?;
        f.write_u32::<LittleEndian>(key_len as u32)?;
        f.write_u32::<LittleEndian>(value_len as u32)?;
        f.write_all(&mut data_tmp)?;
        

        Ok(current_position)
    }

    #[inline]
    pub fn update(&mut self, key: &ByteStr, value: &ByteStr) -> io::Result<()> {
        self.insert(key, value)
    }

}